from pathlib import Path

import matplotlib.pyplot as plt

from pycatalicism.calc.plotter import Plotter
from pycatalicism.calc.plotterexception import PlotterException
from pycatalicism.calc.conversion import Conversion
from pycatalicism.calc.selectivity import Selectivity
from pycatalicism.logging_decorator import Logging

class COOxidationPlotter(Plotter):
    """
    Class for plotting CO oxidation conversion data and exporting resulting plots to file.
    """

    @Logging
    def __init__(self):
        """
        Registers logger with instances of this class which can be accessed via self.logger instance variable
        """

    def plot(self, conversion:Conversion, selectivity:Selectivity|None, show_plot:bool=False, output_plot_path:Path|None=None, plot_title:str|None=None):
        """
        Main interface of this class. Plots conversion vs. temperature as line plot. If show_plot is true, shows plots. If output_data_path was provided, exports plots to result.png to provided directory

        parameters
        ----------
        conversion:Conversion
            wrapper for CO conversion at different temperatures data
        selectivity:Selectivity|None
            should be None since selectivity does not make sense for CO oxidation
        show_plot:bool (default:False)
            if True, show plots
        output_data_path:Path|None (default:None)
            path to directory to export data
        """
        fig, ax = plt.subplots()
        sorted_conversion = conversion.get_sorted()
        ax.plot(sorted_conversion.get_temperatures(), sorted_conversion.get_alphas(), marker='o', markersize=5)
        if plot_title:
            ax.set_title(plot_title)
        ax.set_ylim(bottom=-0.1, top=1.1)
        ax.set_xlabel('Temperature, °C')
        ax.set_ylabel('$\mathrm{CO}$ conversion')
        if show_plot:
            self.logger.info(f'Plotting conversion vs. temperature for CO oxidation reaction')
            plt.show()
        if output_plot_path:
            if output_plot_path.exists() and not output_plot_path.is_dir():
                raise PlotterException(f'Output plot path must be a directory')
            if not output_plot_path.exists():
                output_plot_path.mkdir(parents=True)
            self.logger.info(f'Exporting plot of conversion vs. temperature for CO oxidation reaction')
            dpi = 300
            width = 80 / 25.4
            height = 80 / 25.4
            fig.set_dpi(dpi)
            fig.set_figheight(height)
            fig.set_figwidth(width)
            fig.set_tight_layout(True)
            fig.savefig(fname=output_plot_path.joinpath('result.png'))



from pathlib import Path

from pycatalicism.calc.conversion import Conversion
from pycatalicism.calc.selectivity import Selectivity

class Exporter():
    """
    Abstract class. Concrete instances of child classes are intended for exporting calculated by the program data
    """

    def export(self, output_data_path:Path, conversion:Conversion|None, selectivity:Selectivity|None):
        """
        Concrete classes should override this method to export results of calculations.

        parameters
        ----------
        output_data_path:Path
            path to directory to export resulting data
        conversion:Conversion
            wrapper with conversion vs. temperature data
        selectivity:Selectivity|None
            wrapper with selectivity vs. temperature data or None if selectivity does not make sensefor reaction under consideration

        raises
        ------
        exception:NotImplementedError
            if this method is not overriden
        """
        raise NotImplementedError()



from pycatalicism.calc.calculator import Calculator
from pycatalicism.calc.rawdata import RawData
from pycatalicism.calc.conversion import Conversion
from pycatalicism.calc.selectivity import Selectivity
from pycatalicism.logging_decorator import Logging

class CO2HydrogenationCalculator(Calculator):
    """
    Calculates CO2 conversion and CO, alkanes selectivity at different temperatures from parsed data for CO2 hydrogenation reaction
    """

    @Logging
    def __init__(self):
        """
        Registers logger with the object which can be accessed via self.logger instance variable
        """
        super().__init__()

    def calculate_conversion(self, input_data:RawData) -> Conversion:
        """
        Calculates CO2 conversion at different temperatures for CO2 hydrogenation reaction.

        CO2 conversion is calculated as:

        a = ((pi * fi / Ti) * C(CO2)i - (pf * ff / Tf) * C(CO2)f) /  ((pi * fi / Ti) * C(CO2)i)
        where
            C(CO2)i, C(CO2)f - concentrations of CO2 before and after catalytic reactor, respectively, in mol.%
            fi, ff - total gas flow rates before and after catalytic reactor, respectively, in m^3/s
            pi, pf - pressure of gas at point of total gas flow rate measurement before and after catalytic reactor, respectively, in Pa
            Ti, Tf - temperature of gas at point of total gas flow rate measurement before and after catalytic reactor, respectively, in K

        If flow rate measurement data is not provided, conversion is calculated based solely on CO2 concentrations and warning is logged to console in this case.

        parameters
        ----------
        input_data:RawData
            wrapper of parsed data containing CO2 concentrations at different temperatures as well as initial CO2 concentration before reaction started

        returns
        -------
        conversion:Conversion
            wrapper with CO2 conversion at different temperatures data
        """
        self.logger.info(f'Calculating conversion for CO2 hydrogenation reaction')
        temperatures = []
        alphas = []
        for temperature in input_data.get_temperatures():
            T_i = input_data.get_init_amb_temp()
            p_i = input_data.get_init_amb_pres()
            f_i = input_data.get_init_flow()
            T_f = input_data.get_fin_amb_temp(temperature)
            p_f = input_data.get_fin_amb_pres(temperature)
            f_f = input_data.get_fin_flow(temperature)
            C_CO2_i = input_data.get_init_conc('CO2')
            C_CO2_f = input_data.get_conc('CO2', temperature)
            self.logger.debug(f'{temperature = }')
            self.logger.debug(f'{C_CO2_i = }')
            self.logger.debug(f'{C_CO2_f = }')
            if T_i is None or p_i is None or f_i is None or T_f is None or p_f is None or f_f is None:
                self.logger.warning(f'No data about initial and final flow rate found. Calculating results based only on concentrations')
                T_i = 1
                p_i = 1
                f_i = 1
                T_f = 1
                p_f = 1
                f_f = 1
            alpha = ((p_i * f_i / T_i) * C_CO2_i - (p_f * f_f / T_f) * C_CO2_f) / ((p_i * f_i / T_i) * C_CO2_i)
            self.logger.debug(f'{alpha = }')
            temperatures.append(temperature)
            alphas.append(alpha)
        conversion = Conversion(temperatures, alphas, input_data.get_sample_name())
        return conversion

    def calculate_selectivity(self, input_data:RawData) -> Selectivity:
        """
        Calculates selectivity to CO, CH4, C2H6, C3H8, i-C4H10, n-C4H10, i-C5H12, n-C5H12 at different temperatures from compounds concentrations.

        Selectivity to i-th component is calculated as:

        Si = Xi * n / SUM(Xi * n)
        where
            Xi - concentration of ith component in mol.%
            n - stoichiometry coefficient in CO2 hydrogenation reaction
                CO2 + H2 = CO + H2O, n = 1
                CO2 + 4H2 = CH4 + 2H2O, n = 1
                2CO2 + 7H2 = C2H6 + 4H2O, n = 2
                3CO2 + 10H2 = C3H8 + 6H2O, n = 3
                4CO2 + 13H2 = C4H10 + 8H2O, n = 4
                5CO2 + 16H2 = C5H12 + 10H2O, n = 5

        parameters
        ----------
        input_data:RawData
            wrapper with concentrations of reaction product compounds at different temperatures

        returns
        -------
        selectivity:Selectivity
            wrapper with selectivities to corresponding compounds at different temperatures
        """
        self.logger.info(f'Calculating selectivities for CO2 hydrogenation reaction')
        temperatures = []
        s_list = []
        for temperature in input_data.get_temperatures():
            self.logger.debug(f'{temperature = }')
            c_tot = 0
            s_dict = {}
            for compound in ['CO', 'CH4', 'C2H6', 'C3H8', 'i-C4H10', 'n-C4H10', 'i-C5H12', 'n-C5H12']:
                n_str = compound[compound.find('C')+1]
                n = int(n_str) if n_str.isdecimal() else 1
                s_dict[compound] = input_data.get_conc(compound, temperature) * n
                c_tot = c_tot + s_dict[compound]
                self.logger.debug(f'{compound = }')
                self.logger.debug(f'{n = }')
                self.logger.debug(f'{input_data.get_conc(compound, temperature) = }')
            self.logger.debug(f'{s_dict = }')
            self.logger.debug(f'{sum(list(s_dict.values())) = }')
            self.logger.debug(f'{c_tot = }')
            if c_tot == 0:
                c_tot = 1
            for key in s_dict:
                s_dict[key] = s_dict[key] / c_tot
            self.logger.debug(f'{s_dict = }')
            self.logger.debug(f'{sum(list(s_dict.values())) = }')
            temperatures.append(temperature)
            s_list.append(s_dict)
        selectivity = Selectivity(temperatures, s_list, input_data.get_sample_name())
        return selectivity



from pathlib import Path

from pycatalicism.calc.exporter import Exporter
from pycatalicism.calc.conversion import Conversion
from pycatalicism.calc.selectivity import Selectivity
from pycatalicism.calc.exporterexception import ExporterException
from pycatalicism.logging_decorator import Logging

class COOxidationExporter(Exporter):
    """
    Class for exporting conversion data for CO oxidation reaction to conversion.dat to a directory provided by user of this class.
    """

    @Logging
    def __init__(self):
        """
        Registers logger with instances of this class which can be accessed via self.logger instance variable.
        """
        super().__init__()

    def export(self, output_data_path:Path, conversion:Conversion, selectivity:Selectivity|None):
        """
        Main interface of this class. Exports conversion data for CO oxidation reaction to conversion.dat to a directory provided by user of this method.

        parameters
        ----------
        output_data_path:Path
            path to directory to export resulting data
        conversion:Conversion
            wrapper of CO conversion at different temperatures
        selectivity:Selectivity|None
            should be None since selectivity does not make sense for CO oxidation reaction

        raises
        ------
        exception:ExporterException
            if ouput_data_path exists and not directory
        """
        if output_data_path.exists() and not output_data_path.is_dir():
            raise ExporterException(f'Data path for exporting data must be a folder')
        self._export_conversion(output_data_path, conversion)

    def _export_conversion(self, output_data_path:Path, conversion:Conversion):
        """
        Export CO conversion data to conversion.dat at output_data_path directory

        parameters
        ----------
        ouput_data_path:Path
            path to directory to export conversion data
        conversion:Conversion
            wrapper of CO conversion at different temperatures
        """
        self.logger.info(f'Exporting conversion vs. temperature data for CO oxidation reaction to "{output_data_path.joinpath("conversion.dat")}"')
        if not output_data_path.exists():
            output_data_path.mkdir(parents=True)
        with output_data_path.joinpath('conversion.dat').open(mode='w') as f:
            sorted_conversion = conversion.get_sorted()
            f.write(str(sorted_conversion))



from pycatalicism.calc.calculator import Calculator
from pycatalicism.calc.rawdata import RawData
from pycatalicism.calc.conversion import Conversion
from pycatalicism.logging_decorator import Logging

class COOxidationCalculator(Calculator):
    """
    Class for calculating CO conversion data.
    """

    @Logging
    def __init__(self):
        """
        Registers logger with instance of this class which can be accessed via self.logger instance variable.
        """
        super().__init__()

    def calculate_conversion(self, input_data:RawData) -> Conversion:
        """
        Main interface to this class. Calculate CO conversion at different temperatures for CO oxidation reaction. Conversion is calculated as:

        a = ((pi * fi / Ti) * C(CO)i - (pf * ff / Tf) * C(CO)f) /  ((pi * fi / Ti) * C(CO)i)
        where
            C(CO)i, C(CO)f - concentrations of CO before and after catalytic reactor, respectively, in mol.%
            fi, ff - total gas flow rates before and after catalytic reactor, respectively, in m^3/s
            pi, pf - pressure of gas at point of total gas flow rate measurement before and after catalytic reactor, respectively, in Pa
            Ti, Tf - temperature of gas at point of total gas flow rate measurement before and after catalytic reactor, respectively, in K

        If flow rate measurement data is not provided, conversion is calculated based solely on CO concentrations and warning is logged to console in this case.

        parameters
        ----------
        input_data:RawData
            wrapper with concentrations and flow rate data

        returns
        -------
        conversion:Conversion
            wrapper with conversion vs. temperature data
        """
        self.logger.info(f'Calculating conversion for CO oxidation reaction')
        temperatures = []
        alphas = []
        for temperature in input_data.get_temperatures():
            T_i = input_data.get_init_amb_temp()
            p_i = input_data.get_init_amb_pres()
            f_i = input_data.get_init_flow()
            T_f = input_data.get_fin_amb_temp(temperature)
            p_f = input_data.get_fin_amb_pres(temperature)
            f_f = input_data.get_fin_flow(temperature)
            C_CO_i = input_data.get_init_conc('CO')
            C_CO_f = input_data.get_conc('CO', temperature)
            self.logger.debug(f'{temperature = }')
            self.logger.debug(f'{C_CO_i = }')
            self.logger.debug(f'{C_CO_f = }')
            if T_i is None or p_i is None or f_i is None or T_f is None or p_f is None or f_f is None:
                self.logger.warning(f'No data about initial and final flow rate found. Calculating results based only on concentrations')
                T_i = 1
                p_i = 1
                f_i = 1
                T_f = 1
                p_f = 1
                f_f = 1
            alpha = ((p_i * f_i / T_i) * C_CO_i - (p_f * f_f / T_f) * C_CO_f) / ((p_i * f_i / T_i) * C_CO_i)
            temperatures.append(temperature)
            alphas.append(alpha)
        conversion = Conversion(temperatures, alphas, input_data.get_sample_name())
        return conversion

    def calculate_selectivity(self, input_data:RawData) -> None:
        """
        Overrides method of superclass. Just returns None since selectivity for CO oxidation reaction does not make sense.
        """
        return None



from pycatalicism.calc.plotter import Plotter
from pycatalicism.calc.co_oxidation_plotter import COOxidationPlotter
from pycatalicism.calc.co2_hydrogenation_plotter import CO2HydrogenationPlotter
from pycatalicism.calc.plotterexception import PlotterException

"""
Factory for creation of resulting data plotters.
"""

def get_plotter(reaction:str) -> Plotter:
    """
    Get plotter for specified reaction

    parameters
    ----------
    reaction:str {co-oxidation|co2-hydrogenation}
        chemical reaction of interest

    returns
    -------
    plotter:Plotter
        plotter for specified reaction

    raises
    ------
    exception:PlotterException
        if unknown reaction provided as parameter
    """
    if reaction == 'co-oxidation':
        return COOxidationPlotter()
    elif reaction == 'co2-hydrogenation':
        return CO2HydrogenationPlotter()
    else:
        raise PlotterException(f'Cannot create plotter for reaction "{reaction}"')



class ChromatographException(Exception):
    """
    Common cromatograph exceptions
    """

class ChromatographStateException(ChromatographException):
    """
    Exceptions related to chromatograph state
    """



# mass flow controllers calibrations
# see config.py for details
calibrations = [
                0, # He mass flow controller
                2, # CO2/O2 mass flow controller
                2  # CH4/H2/CO mass flow controller
                ]
# flow rates of gases during activation step
activation_flow_rates = [
                        0, # He mass flow controller
                        0, # CO2/O2 mass flow controller
                        36 # CH4/H2/CO mass flow controller
                        ]
# temperature of activation in °C
activation_temperature = 300
# activation duration in minutes
activation_duration = 60
# after the activation furnace is turned off and below this temperature gas flow rates will be changed to post_flow_rates values
post_temperature = 100
# flow rates of gases after the activation is over
post_flow_rates = [
                0, # He mass flow controller
                0, # CO2/O2 mass flow controller
                3  # CH4/H2/CO mass flow controller
                ]



from pathlib import Path

from pycatalicism.calc.parser import Parser
from pycatalicism.calc.rawdata import RawData
from pycatalicism.calc.parserexception import ParserException
from pycatalicism.logging_decorator import Logging

class ChromatecCrystalCompositionCopyPasteParser(Parser):
    """
    Class for parsing data obtained by simple copy-paste from chromatec analytics software and adding relevant data to resulting file.

    Data for this parser must be in the following format:

    Температура<tab><temperature>
    <br>
    Название<tab>Время, мин<tab>Детектор<tab>Концентрация<tab>Ед, измерения<tab>Площадь<tab>Высота
    <compound-name><tab><retention-time><tab><detector-name><tab><compound-concentration><tab><concentration-units><tab><peak-area><tab><peak-height>
    [<br>
    Темп. (газовые часы)<tab><flow-temperature>
    Давление (газовые часы)<tab><flow-pressure>
    Поток<tab><flow-rate>]
    """

    @Logging
    def __init__(self):
        """
        Registers logger to the object which can be used by self.logger instance variable.
        """
        super().__init__()

    def parse_data(self, input_data_path:Path, initial_data_path:Path, sample_name:str|None) -> RawData:
        """
        Main interface to the class. Parses concentration, temperature and, if present, flow rate data from data files. Data must be in the following format:

        Температура<tab><temperature>
        <br>
        Название<tab>Время, мин<tab>Детектор<tab>Концентрация<tab>Ед, измерения<tab>Площадь<tab>Высота
        <compound-name><tab><retention-time><tab><detector-name><tab><compound-concentration><tab><concentration-units><tab><peak-area><tab><peak-height>
        [<br>
        Темп. (газовые часы)<tab><flow-temperature>
        Давление (газовые часы)<tab><flow-pressure>
        Поток<tab><flow-rate>]

        If format in a file is wrong that file is ignored and warning is logged via self.logger

        parameters
        ----------
        input_data_path:Path
            path to folder with initial data files
        initial_data_path:Path
            path to file with initial data (i.e. when reaction did not occured)
        sample_name:str|None
            name of sample used as label for plotting

        returns
        -------
        raw_data:RawData
            wrapper with parsed data

        raises
        ------
        exception:ParserException
            if initial_data_path is not file or if input_data_path is not directory
        """
        if not initial_data_path.is_file():
            raise ParserException(f'initial data path {initial_data_path} must be a file')
        if not input_data_path.is_dir():
            raise ParserException(f'input data path {input_data_path} must be a directory')
        _, Cs_i, Ta_i, Pa_i, f_i = self._parse_file(initial_data_path)
        flow_is_measured = Ta_i and Pa_i and f_i
        Ts = []
        Cs_f = []
        Ta_f = [] if flow_is_measured else None
        Pa_f = [] if flow_is_measured else None
        f_f = [] if flow_is_measured else None
        for file in input_data_path.iterdir():
            if file == initial_data_path:
                continue
            if file.is_dir():
                self.logger.warning(f'Found directory {file} in input data path')
                continue
            try:
                T, C, Ta, Pa, f = self._parse_file(file)
            except ParserException:
                self.logger.warning(f'Wrong data format in file {file}. Skipping.')
                continue
            Ts.append(T)
            Cs_f.append(C)
            if Ta_f is not None and Pa_f is not None and f_f is not None:
                Ta_f.append(Ta)
                Pa_f.append(Pa)
                f_f.append(f)
        rawdata = RawData(temperatures=Ts, initial_concentrations=Cs_i, concentrations=Cs_f, initial_ambient_temperature=Ta_i, initial_ambient_pressure=Pa_i, initial_flow=f_i, final_ambient_temperatures=Ta_f, final_ambient_pressures=Pa_f, final_flows=f_f, sample_name=sample_name)
        return rawdata

    def _parse_file(self, path:Path) -> tuple[float,dict[str,float],float|None,float|None,float|None]:
        """
        Parse single file with data. Data in a file must be in a following format:

        Температура<tab><temperature>
        <br>
        Название<tab>Время, мин<tab>Детектор<tab>Концентрация<tab>Ед, измерения<tab>Площадь<tab>Высота
        <compound-name><tab><retention-time><tab><detector-name><tab><compound-concentration><tab><concentration-units><tab><peak-area><tab><peak-height>
        [<br>
        Темп. (газовые часы)<tab><flow-temperature>
        Давление (газовые часы)<tab><flow-pressure>
        Поток<tab><flow-rate>]

        All commas in a file is replaced with dots before parsing, so decimal separator may be both "," and "."

        parameters
        ----------
        path:Path
            path to file with data

        returns
        -------
        (T, C, Ta, Pa, f):tuple
            T:float
                temperature at which reaction taken place
            C:dict[str:float]
                dictionary of compounds and their concentrations in mol.%
            Ta:float|None
                temperature at which measurement of flow rate was done or None if not present in a file (NB: units)
            Pa:float|None
                pressure at which measurement of flow rate was done or None if not present in a file (NB: units)
            f:float|None
                total gas flow rate or None if not present in a file (NB: units)

        raises
        ------
        exception:ParserException
            if data format is wrong
        """
        try:
            file_contents = self._replace_commas_with_dots(path)
        except UnicodeDecodeError:
            raise ParserException(f'Non unicode file {path}')
        T = None
        C = {}
        Ta = None
        Pa = None
        f = None
        lines = file_contents.split(sep='\n')
        while lines:
            line = lines.pop(0)
            words = line.split(sep='\t')
            self.logger.debug(f'processing line: "{line}"')
            if 'Температура' in words:
                T = float(words[1])
            if 'Название' in words and 'Концентрация' in words:
                compound_index = words.index('Название')
                concentration_index = words.index('Концентрация')
                while True:
                    if lines:
                        line = lines.pop(0)
                        words = line.split(sep='\t')
                        if line == '':
                            break
                        self.logger.debug(f'processing line: "{line}"')
                        compound = words[compound_index]
                        concentration = words[concentration_index]
                        C[compound] = float(concentration)
                    else:
                        break
            if 'Темп. (газовые часы)' in words:
                Ta = float(line.split(sep='\t')[1])
            if 'Давление (газовые часы)' in words:
                Pa = float(line.split(sep='\t')[1])
            if 'Поток' in words:
                f = float(line.split(sep='\t')[1])
        if T is None or len(C) == 0:
            raise ParserException(f'Wrong data format in file {path}')
        return (T, C, Ta, Pa, f)

    def _replace_commas_with_dots(self, path:Path) -> str:
        """
        replace all commas with dots in a file

        parameters
        ----------
            path:Path
                path to file for processing

        returns
        -------
            new_contents:str
                string with file contents in which commas were replaced with dots
        """
        with path.open(mode='r') as file:
            contents = file.read()
            new_contents = contents.replace(',', '.')
        return new_contents



# mass flow controllers calibrations
# see config.py for details
calibrations = [
                0, # He mass flow controller
                2, # CO2/O2 mass flow controller
                4, # CH4/H2/CO mass flow controller
                ]
# flow rates of gases for measurement
flow_rates = [
            30, # He mass flow controller
            4, # CO2/O2 mass flow controller
            2, # CH4/H2/CO mass flow controller
            ]
# chromatograph instrumental method to use for measurement
chromatograph_method = 'co-oxidation'
# minimal isothermal dwell time in minutes at each temperature to reach steady-state conditions
isothermal = 30
# list of measurement temperatures
temperatures = []
# name of sample
sample_name = 
# operator's name
operator = 



import functools
import inspect
import logging
import sys
from typing import Callable

import pycatalicism.config as config

class Logging:
    """
    Class used to register and configure logger with objects or modules. In order to use it one should decorate any method in module or __init__ method in object class with Logging class. The registered logger is available as "logger" or "self.logger" instance variables
    """

    def __init__(self, func:Callable):
        """
        Assign parameter to instance variable, get logging levels from config file placed in the same directory as this module.

        parameters
        ----------
        func:Callable
            function to be decorated with this class
        """
        functools.update_wrapper(self, func)
        self.func = func
        self.logging_levels = config.logging_levels

    def __get__(self, obj:object, type=None):
        """
        When this class is initialized at import time, it does not know whether func is function or method. This method is used to overcome the problem (see https://stackoverflow.com/questions/2366713/can-a-decorator-of-an-instance-method-access-the-class/48491028#48491028)

        parameters
        ----------
        parameters are overriden from standard python __get__ method

        returns
        -------
        self.__class_(func)
        """
        func = self.func.__get__(obj, type)
        return self.__class__(func)

    def __call__(self, *args, **kwargs):
        """
        To make class callable, this method is overriden. If self.func is function, add configured logger to its module's namespace, else if self.func is method add configured logger to its object namespace.

        parameters
        ----------
        args
            positional arguments to self.func
        kwargs
            keyword arguments to self.func

        returns
        -------
        self.func(*args, **kwargs)
            return value, returned by self.func
        """
        if inspect.isfunction(self.func):
            module = sys.modules[self.func.__module__]
            if 'logger' not in module.__dict__:
                logger = logging.getLogger(module.__name__)
                self._configure_logger(logger, self.logging_levels[module.__name__])
                module.__dict__['logger'] = logger
        elif inspect.ismethod(self.func):
            obj = self.func.__self__
            if 'logger' not in obj.__dict__:
                logger = logging.getLogger(obj.__class__.__name__)
                self._configure_logger(logger, self.logging_levels[obj.__class__.__name__])
                obj.__dict__['logger'] = logger
        else:
            raise Exception(f'Cannot decorate function {self.func.__name__}')
        return self.func(*args, **kwargs)

    def _configure_logger(self, logger:logging.Logger, level:int):
        """
        Set level to logger, add StreamHandler (will log to console) and formatter

        parameters
        ----------
        logger:Logger
            logger to be configured
        level:int
            logging level
        """
        logger.setLevel(level)
        logger.propagate = False

        ch = logging.StreamHandler()
        ch.setLevel(level)

        formatter = logging.Formatter(fmt='[%(asctime)s] %(name)s.%(funcName)s: %(levelname)s: %(message)s', datefmt='%d.%m.%Y %H:%M:%S')

        ch.setFormatter(formatter)

        logger.addHandler(ch)



from pathlib import Path

from pycatalicism.calc.rawdata import RawData

class Parser():
    """
    Abstract class for parsers of initial data.
    """

    def parse_data(self, input_data_path:Path, initial_data_path:Path, sample_name:str|None) -> RawData:
        """
        Methods of concrete classes should override this method.

        parameters
        ----------
        input_data_path:Path
            path to directory with data files
        initial_data_path:Path
            path to file with initial (i.e. before catalytic reaction occured) dad
        sample_name:str|None
            name of sample used as label for plotting

        returns
        -------
        raw_data:RawData
            wrapper with parsed data

        raises
        ------
        exception:NotImplementedError
            if this method is not overriden
        """
        raise NotImplementedError()



class FurnaceProtocolException(Exception):
    """
    Exceptions related to connection protocol.
    """

class FurnaceConnectionException(Exception):
    """
    Exceptions related to connection issues.
    """

class FurnaceStateException(Exception):
    """
    Exceptions related to wrong furnace state.
    """



import time

import pycatalicism.chromatograph.chromatograph_logging as chromatograph_logging
from pycatalicism.chromatograph.chromatec_control_panel_modbus import ChromatecControlPanelModbus
from pycatalicism.chromatograph.chromatec_control_panel_modbus import ConnectionStatus
from pycatalicism.chromatograph.chromatec_control_panel_modbus import WorkingStatus
from pycatalicism.chromatograph.chromatec_control_panel_modbus import ChromatographCommand
from pycatalicism.chromatograph.chromatec_control_panel_modbus import ApplicationCommand
from pycatalicism.chromatograph.chromatec_analytic_modbus import ChromatogramPurpose
from pycatalicism.chromatograph.chromatec_analytic_modbus import ChromatecAnalyticModbus
from pycatalicism.chromatograph.chromatograph_exceptions import ChromatographException
from pycatalicism.chromatograph.chromatograph_exceptions import ChromatographStateException

class ChromatecCrystal5000():
    """
    Class represents chromatec crystal 5000 chromatograph. Communication is done via modbus protocol.
    """

    def __init__(self, control_panel:ChromatecControlPanelModbus, analytic:ChromatecAnalyticModbus, methods:dict[str, int]):
        """
        Initializes instance variables, registers logger.
        """
        self._control_panel = control_panel
        self._analytic = analytic
        self._methods = methods
        self._logger = chromatograph_logging.get_logger(self.__class__.__name__)
        self._connected = False

    def connect(self):
        """
        Connect to chromatograph. If chromatec control panel is not up, start control panel, connection is established automatically in this case. If control panel is up, but chromatograph is disconnected, establish connection. Do nothing otherwise.

        raises
        ------
        ChromatographException
            if unknown connection status was get from chromatograph
        """
        connection_status = self._control_panel.get_connection_status()
        if connection_status is ConnectionStatus.CP_OFF_NOT_CONNECTED:
            self._logger.info('Starting control panel application')
            self._control_panel.send_application_command(ApplicationCommand.START_CONTROL_PANEL)
            self._logger.info('Waiting until control panel is up...')
            while True:
                connection_status = self._control_panel.get_connection_status()
                self._logger.debug(f'{connection_status = }')
                if connection_status is ConnectionStatus.CP_ON_CONNECTED:
                    break
                time.sleep(10)
            self._connected = True
            self._logger.info('Control panel is UP. Connection established.')
        elif connection_status is ConnectionStatus.CP_ON_NOT_CONNECTED:
            self._logger.info('Connecting to chromatograph')
            self._control_panel.send_chromatograph_command(ChromatographCommand.CONNECT_CHROMATOGRAPH)
            self._logger.info('Waiting until connection is established...')
            while True:
                connection_status = self._control_panel.get_connection_status()
                self._logger.debug(f'{connection_status = }')
                if connection_status is ConnectionStatus.CP_ON_CONNECTED:
                    break
                time.sleep(10)
            self._connected = True
            self._logger.info('Connection established')
        elif connection_status is ConnectionStatus.CP_ON_CONNECTED:
            self._logger.info('Chromatograph connected already')
            self._connected = True
        else:
            raise ChromatographException(f'Unknown connection status: {connection_status}')

    def set_method(self, method:str):
        """
        Set chromatograph instrument method. Chromatograph start to prepare itself for analysis accordingly.

        parameters
        ----------
        method:str
            method to send to chromatograph

        raises
        ------
        ChromatographStateException
            if connection to chromatograph is not established or analysis is in progress now
        """
        working_status = self._control_panel.get_current_working_status()
        if not self._connected:
            raise ChromatographStateException('Connect to chromatograph first!')
        if working_status is WorkingStatus.ANALYSIS:
            raise ChromatographStateException('Analysis is in progress!')
        self._logger.info(f'Setting method to {method}')
        self._control_panel.set_instrument_method(self._methods[method])

    def is_ready_for_analysis(self) -> bool:
        """
        Check if chromatograph is ready for analysis.

        returns
        -------
        is_ready_for_analysis:bool
            True if chromatograph is ready for analysis

        raises
        ------
        ChromatographStateException
            if connection to chromatograph is not established
        """
        working_status = self._control_panel.get_current_working_status()
        if not self._connected:
            raise ChromatographStateException('Connect to chromatograph first!')
        self._logger.info('Checking if chromatograph is ready for analysis')
        is_ready_for_analysis = working_status is WorkingStatus.READY_FOR_ANALYSIS
        self._logger.info(f'Chromatograph is ready for analysis: {is_ready_for_analysis}')
        return is_ready_for_analysis

    def start_analysis(self):
        """
        Start analysis.

        raises
        ------
        ChromatographStateException
            if chromatograph is not connected or instrumental method was not started yet or chromatograph is not ready to start analysis
        """
        working_status = self._control_panel.get_current_working_status()
        if not self._connected:
            raise ChromatographStateException('Connect to chromatograph first!')
        if working_status is WorkingStatus.NULL:
            raise ChromatographStateException('Start some instrumental method first!')
        if working_status is not WorkingStatus.READY_FOR_ANALYSIS:
            raise ChromatographStateException('Chromatograph is not ready to start analysis')
        else:
            self._logger.info('Starting analysis')
            self._control_panel.send_chromatograph_command(ChromatographCommand.START_ANALYSIS)

    def set_passport(self, name:str, volume:float, dilution:float, purpose:ChromatogramPurpose, operator:str, column:str, lab_name:str):
        """
        Set passport values for chromatogram. Method should be called after the analysis is finished otherwise previous chromatogram's passport will be changed.

        parameters
        ----------
        name:str
            name of chromatogram
        volume:float
            volume of sample
        dilution:float
            dilution of sample
        purpose:ChromatogramPurpose
            analysis or graduation
        operator:str
            name of operator
        column:str
            name of column
        lab_name:str
            name of laboratory

        raises
        ------
        ChromatographStateException
            if chromatograph is not connected or method was not started yet or analysis is in progress
        """
        working_status = self._control_panel.get_current_working_status()
        if not self._connected:
            raise ChromatographStateException('Connect chromatograph first!')
        if working_status is WorkingStatus.NULL:
            raise ChromatographStateException('Start some instrumental method first!')
        if working_status is WorkingStatus.ANALYSIS:
            raise ChromatographStateException('Analysis is in progress, cannot set passport for currently running chromatogram, wait until analysis is over!')
        self._logger.info(f'Setting passport values to: [name:{name}, volume:{volume}, dilution:{dilution}, purpose:{purpose}, operator:{operator}, column:{column}, lab_name:{lab_name}]')
        self._analytic.set_sample_name(name)
        self._analytic.set_sample_volume(volume)
        self._analytic.set_sample_dilution(dilution)
        self._analytic.set_chromatogram_purpose(purpose)
        self._analytic.set_operator(operator)
        self._analytic.set_column(column)
        self._analytic.set_lab_name(lab_name)

    def get_working_status(self) -> WorkingStatus:
        """
        Get working status of chromatograph.

        returns
        -------
        working_status:WorkingStatus
            chromatograph working status

        raises
        ------
        ChromatographStateException
            if chromatograph is not connected
        """
        if not self._connected:
            raise ChromatographStateException('Connect chromatograph first!')
        self._logger.info('Getting working status.')
        working_status = self._control_panel.get_current_working_status()
        self._logger.info(f'{working_status = }')
        return working_status



from enum import Enum

from pymodbus.client.sync import ModbusTcpClient

import pycatalicism.chromatograph.chromatograph_logging as chromatograph_logging
import pycatalicism.chromatograph.modbus_converter as convert

class ChromatogramPurpose(Enum):
    """
    Prupose of chromatogram as is written in passport
    """
    ANALYSIS = 0
    GRADUATION = 1

class ChromatecAnalyticModbus():
    """
    Class represents modbus protocol for connection with chromatec analytic modbus slave
    """

    def __init__(self, modbus_id:int, sample_name_holding_address:int, chromatogram_purpose_holding_address:int, sample_volume_holding_address:int, sample_dilution_holding_address:int, operator_holding_address:int, column_holding_address:int, lab_name_holding_address:int):
        """
        Initializes private instance variables. Registers logger.

        parameters
        ----------
        modbus_id:int
            modbus slave id of analytic software
        sample_name_holding_address:int
            modbus address for chromatogram name as written in the passport of chromatogram
        chromatogram_purpose_holding_address:int
            modbus address for chromatogram purpose as written in the passport of chromatogram
        sample_volume_holding_address:int
            modbus address for sample volume as written in the passport of chromatogram
        sample_dilution_holding_address:int
            modbus address for sample dilution as written in the passport of chromatogram
        operator_holding_address:int
            modbus address for operator as written in the passport of chromatogram
        column_holding_address:int
            modbus address for column as written in the passport of chromatogram
        lab_name_holding_address:int
            modbus address for lab name as written in the passport of chromatogram
        """
        self._modbus_id = modbus_id
        self._sample_name_holding_address = sample_name_holding_address
        self._chromatogram_purpose_holding_address = chromatogram_purpose_holding_address
        self._sample_volume_holding_address = sample_volume_holding_address
        self._sample_dilution_holding_address = sample_dilution_holding_address
        self._operator_holding_address = operator_holding_address
        self._column_holding_address = column_holding_address
        self._lab_name_holding_address = lab_name_holding_address
        self._modbus_client = ModbusTcpClient()
        self._logger = chromatograph_logging.get_logger(self.__class__.__name__)

    def set_sample_name(self, name:str):
        """
        Set name of sample in chromatogram's passport.

        parameters
        ----------
        name:str
            name of sample
        """
        self._logger.debug(f'Setting chromatogram name to {name}')
        name_bytes = convert.string_to_bytes(name)
        self._modbus_client.write_registers(address=self._sample_name_holding_address, values=name_bytes, unit=self._modbus_id)

    def set_chromatogram_purpose(self, purpose:ChromatogramPurpose):
        """
        Set purpose of chromatogram in chromatogram's passport.

        parameters
        ----------
        purpose:ChromatogramPurpose
            one of the constants defined in ChromatogramPurpose enum
        """
        self._logger.debug(f'Setting chromatogram purpose to {purpose}')
        purpose_bytes = convert.int_to_bytes(purpose.value)
        self._modbus_client.write_registers(address=self._chromatogram_purpose_holding_address, values=purpose_bytes, unit=self._modbus_id)

    def set_sample_volume(self, volume:float):
        """
        Set sample volume in chromatogram's passport.

        parameters
        ----------
        volume:float
            sample volume
        """
        self._logger.debug(f'Setting sample volume to {volume}')
        volume_bytes = convert.double_to_bytes(volume)
        self._modbus_client.write_registers(address=self._sample_volume_holding_address, values=volume_bytes, unit=self._modbus_id)

    def set_sample_dilution(self, dilution:float):
        """
        Set sample dilution in chromatogram's passport.

        parameters
        ----------
        dilution:float
            sample dilution
        """
        self._logger.debug(f'Setting sample dilution to {dilution}')
        dilution_bytes = convert.double_to_bytes(dilution)
        self._modbus_client.write_registers(address=self._sample_dilution_holding_address, values=dilution_bytes, unit=self._modbus_id)

    def set_operator(self, operator:str):
        """
        Set operator in chromatogram's passport.

        parameters
        ----------
        operator:str
            operator's name
        """
        self._logger.debug(f'Setting operator to {operator}')
        operator_bytes = convert.string_to_bytes(operator)
        self._modbus_client.write_registers(address=self._operator_holding_address, values=operator_bytes, unit=self._modbus_id)

    def set_column(self, column:str):
        """
        Set column in chromatogram's passport.

        parameters
        ----------
        column:str
            column's name
        """
        self._logger.debug(f'Setting column to {column}')
        column_bytes = convert.string_to_bytes(column)
        self._modbus_client.write_registers(address=self._column_holding_address, values=column_bytes, unit=self._modbus_id)

    def set_lab_name(self, name:str):
        """
        Set laboratory name in chromatogram's passport

        parameters
        ----------
        name:str
            laboratory name
        """
        self._logger.debug(f'Setting laboratory name to {name}')
        name_bytes = convert.string_to_bytes(name)
        self._modbus_client.write_registers(address=self._lab_name_holding_address, values=name_bytes, unit=self._modbus_id)

    def get_lab_name(self) -> str:
        """
        Get laboratory name from chromatogram's passport.

        returns
        -------
        name:str
            laboratory name
        """
        self._logger.debug('Getting laboratory name')
        response = self._modbus_client.read_holding_registers(address=self._lab_name_holding_address, count=15, unit=self._modbus_id)
        name = convert.bytes_to_string(response.registers)
        self._logger.log(5, f'{name = }')
        return name



from pathlib import Path

import matplotlib.pyplot as plt
import matplotlib
import matplotlib.axes
import numpy as np

from pycatalicism.calc.plotter import Plotter
from pycatalicism.calc.plotterexception import PlotterException
from pycatalicism.calc.conversion import Conversion
from pycatalicism.calc.selectivity import Selectivity
from pycatalicism.logging_decorator import Logging

class CO2HydrogenationPlotter(Plotter):
    """
    Class for plotting CO2 hydrogenation conversion and selectivity data and exporting resulting plots to file.
    """

    @Logging
    def __init__(self):
        """
        Registers logger with instances of this class which can be accessed via self.logger instance variable
        """
        super().__init__()

    def plot(self, conversion:Conversion|None, selectivity:Selectivity|None, show_plot:bool=False, output_plot_path:Path|None=None, plot_title:str|None=None):
        """
        Main interface of this class. Plots conversion vs. temperature as line plot and selectivities vs. temperature as bar plot. If show_plot is true, shows plots. If output_data_path was provided, exports plots to result.png to provided directory

        parameters
        ----------
        conversion:Conversion
            wrapper for CO2 conversion at different temperatures data
        selectivity:Selectivity
            wrapper for selectivities to different compounds at different temperatures data
        show_plot:bool (default:False)
            if True, show plots
        output_data_path:Path|None (default:None)
            path to directory to export data

        raises
        ------
        PlotterException
            if conversion and selectivities are None
        """
        if conversion and selectivity:
            fig, (ax_conversion, ax_selectivity) = plt.subplots(nrows=1, ncols=2)
            ax_conversion = self._plot_conversion(ax_conversion, conversion)
            ax_selectivity = self._plot_selectivity(ax_selectivity, selectivity)
        elif conversion and not selectivity:
            fig, ax_conversion = plt.subplots()
            ax_conversion = self._plot_conversion(ax_conversion, conversion)
        elif selectivity and not conversion:
            fig, ax_selectivity = plt.subplots()
            ax_selectivity = self._plot_selectivity(ax_selectivity, selectivity)
        else:
            raise PlotterException('Nothing to plot')
        if plot_title:
            fig.suptitle(plot_title)
        if show_plot:
            self.logger.info(f'Plotting conversion vs. temperature for CO2 hydrogenation reaction')
            plt.show()
        if output_plot_path:
            if output_plot_path.exists() and not output_plot_path.is_dir():
                raise PlotterException(f'Output plot path must be a directory')
            if not output_plot_path.exists():
                output_plot_path.mkdir(parents=True)
            self.logger.info(f'Exporting plot of conversion vs. temperature for CO oxidation reaction')
            dpi = 300
            width = 160 / 25.4
            height = 80 / 25.4
            fig.set_dpi(dpi)
            fig.set_figheight(height)
            fig.set_figwidth(width)
            fig.set_tight_layout(True)
            fig.savefig(fname=output_plot_path.joinpath('result.png'))

    def _plot_conversion(self, ax:matplotlib.axes.Axes, conversion:Conversion) -> matplotlib.axes.Axes:
        """
        Plot CO2 conversion vs. temperature plot as line plot.

        parameters
        ----------
        ax:Axes
            axes to plot to
        conversion:Conversion
            wrapper to CO2 conversion at different temperatures data

        returns
        -------
        ax:Axes
            axes with plotted data
        """
        sorted_conversion = conversion.get_sorted()
        ax.plot(sorted_conversion.get_temperatures(), sorted_conversion.get_alphas(), marker='o', markersize=5)
        _max = sorted_conversion.get_alphas().max()
        _min = sorted_conversion.get_alphas().min()
        delta = _max - _min
        ax.set_ylim(bottom=_min - 0.1 * delta, top=_max + 0.1 * delta)
        ax.set_xlabel('Temperature, °C')
        ax.set_ylabel('$\mathrm{CO_2}$ conversion')
        return ax

    def _plot_selectivity(self, ax:matplotlib.axes.Axes, selectivity:Selectivity) -> matplotlib.axes.Axes:
        """
        Plot selectivities to different compounds at different temperatures as bar plot

        parameters
        ----------
        ax:Axes
            axes to plot to
        selectivity:Selectivity
            wrapper with selectivity data

        returns
        -------
        ax:Axes
            axes with plotted data
        """
        sorted_selectivity = selectivity.get_sorted()
        self.logger.debug(f'{str(sorted_selectivity) = }')
        s_dict = {}
        for temperature in sorted_selectivity.get_temperatures():
            self.logger.debug(f'{temperature = }')
            for compound in sorted_selectivity.get_selectivities_at(temperature):
                self.logger.debug(f'{compound = }')
                self.logger.debug(f'{s_dict = }')
                if compound in s_dict:
                    self.logger.debug(f'{sorted_selectivity.get_selectivity(compound, temperature) = }')
                    s_dict[compound].append(sorted_selectivity.get_selectivity(compound, temperature))
                else:
                    self.logger.debug(f'{sorted_selectivity.get_selectivity(compound, temperature) = }')
                    s_dict[compound] = [sorted_selectivity.get_selectivity(compound, temperature)]
        compounds = list(s_dict)
        bottom = 0
        for i in range(len(compounds)):
            self.logger.debug(f'{compounds[i] = }')
            self.logger.debug(f'{sorted_selectivity.get_temperatures() = }')
            self.logger.debug(f'{s_dict[compounds[i]] = }')
            self.logger.debug(f'{bottom = }')
            if not np.all(np.array(s_dict[compounds[i]]) == 0):
                ax.bar(sorted_selectivity.get_temperatures(), s_dict[compounds[i]], bottom=bottom, width=5, label=compounds[i])
                bottom = bottom + np.array(s_dict[compounds[i]])
        ax.set_xlabel('Temperature, °C')
        ax.set_ylabel('Selectivity')
        ax.legend(loc='upper left', bbox_to_anchor=(1, 1))
        return ax



from pycatalicism.calc.calculator import Calculator
from pycatalicism.calc.cooxidationcalculator import COOxidationCalculator
from pycatalicism.calc.co2hydrogenationcalculator import CO2HydrogenationCalculator
from pycatalicism.calc.co2hydrogenationproductsbasiscalculator import CO2HydrogenationProductsBasisCalculator
from pycatalicism.calc.calculatorexception import CalculatorException

"""
Factory to create calculator for proper reaction.
"""

def get_calculator(reaction:str, products_basis:bool) -> Calculator:
    """
    Create calculator for reaction to calculate conversion, selectivity or activity vs. temperature data.

    parameters
    ----------
    reaction:str {co-oxidation|co2-hydrogenation}
        Chemical reaction for which to calculate results
    products_basis:bool
        If True, return calculator, which calculates conversion based on products composition

    returns
    -------
    calculator:Calculator
        Calculator to use for catalyst characteristics calculation

    raises
    ------
    exception:Exception
        if reaction is not known
    """
    if reaction == 'co-oxidation':
        return COOxidationCalculator()
    elif reaction == 'co2-hydrogenation':
        if products_basis:
            return CO2HydrogenationProductsBasisCalculator()
        else:
            return CO2HydrogenationCalculator()
    else:
        raise CalculatorException(f'Cannot create calculator for reaction {reaction}')



from pathlib import Path

from pycatalicism.calc.conversion import Conversion
from pycatalicism.calc.selectivity import Selectivity

class Plotter():
    """
    Abstract class for plotting resulting data
    """

    def plot(self, conversion:Conversion, selectivity:Selectivity|None, show_plot:bool=False, output_plot_path:Path|None=None, plot_title:str|None=None):
        """
        Method should be overriden by concrete classes. Plots conversion, selectivity vs. temperature plots, shows them and exports to file.

        parameters
        ----------
        conversion:Conversion
            wrapper for conversion data
        selectivity:Selectivity|None
            wrapper for selectivity data or None if selectivity does not make sense for reaction
        show_plot:bool (default:False)
            show plot if True
        output_plot_path:Path|None (default:None)
            path to directory to export resulting plots or None if export is not needed

        raises
        ------
        NotImplementedError
            if method is not overriden
        """
        raise NotImplementedError()



from pycatalicism.calc.parser import Parser
from pycatalicism.calc.parserexception import ParserException
from pycatalicism.calc.chromatec_crystal_composition_copy_paste_parser import ChromatecCrystalCompositionCopyPasteParser

"""
Factory for creating parser for specific data format.
"""

def get_parser(parser_type:str) -> Parser:
    """
    Get parser for specified data format.

    parameters
    ----------
    parser_type:str {chromatec-crystal-composition-copy-paste}
        parser type representing certain data format

    raises
    ------
    exception:ParserException
        if parser type is not known
    """
    if parser_type == 'chromatec-crystal-composition-copy-paste':
        return ChromatecCrystalCompositionCopyPasteParser()
    else:
        raise ParserException(f'cannot create parser for {parser_type}')



from pycatalicism.calc.exporter import Exporter
from pycatalicism.calc.co2_hydrogenation_exporter import CO2HydrogenationExporter
from pycatalicism.calc.co_oxidation_exporter import COOxidationExporter
from pycatalicism.calc.exporterexception import ExporterException

"""
Factory for creating exporters for specified reaction.
"""

def get_exporter(reaction:str) -> Exporter:
    """
    Get exporter for specified reaction.

    parameters
    ----------
    reaction:str {co-oxidation|co2-hydrogenration}
        chemical reaction to export calculated results

    returns
    -------
    exporter:Exporter
        concrete exporter object for specified reaction

    raises
    ------
    exception:ExporterException
        if reaction is not known
    """
    if reaction == 'co-oxidation':
        return COOxidationExporter()
    elif reaction == 'co2-hydrogenation':
        return CO2HydrogenationExporter()
    else:
        raise ExporterException(f'Cannot create exporter for reaction "{reaction}"')



import pycatalicism.furnace.furnace_logging as furnace_logging
from pycatalicism.furnace.owen_protocol import OwenProtocol
from pycatalicism.furnace.furnace_exceptions import FurnaceConnectionException, FurnaceStateException

class OwenTPM101():
    """
    Class represents simplified version of Owen TPM101 temperature controller. Far less parameters can be used in this class in comparison to the real controller.
    """

    def __init__(self, device_name:str, owen_protocol:OwenProtocol):
        """
        Assign parameters to instance variables, initialize connection flag and register logger.

        parameters
        ----------
        device_name:str
            The name of the controller which is returned via serial protocol.
        owen_protocol:OwenProtocol
            Protocol used for communication over serial port.
        """
        self._connected = False
        self._owen_protocol = owen_protocol
        self._device_name = device_name
        self._logger = furnace_logging.get_logger(self.__class__.__name__)

    def connect(self):
        """
        Connect to the device. Check if device name returned by the controller is the same as one provided at the initialization. This method must be run before any other method of this class.

        raises
        ------
        FurnaceConnectionException
            if device_name does not match
        """
        self._logger.info('Connecting to temperature controller.')
        device_name = self._owen_protocol.request_string(parameter='dev')
        if self._device_name != device_name:
            raise FurnaceConnectionException('Cannot connect to furnace controller!')
        self._connected = True

    def set_temperature(self, temperature:float):
        """
        Set temperature to the specified value.

        parameters
        ----------
        temperature:float
            temperature in °C

        raises
        ------
        FurnaceStateException
            if connect method was not called
        """
        if not self._connected:
            raise FurnaceStateException('Connect to furnace controller first!')
        self._logger.info(f'Setting temperature to {temperature}°C')
        self._owen_protocol.send_PIC(parameter='sp', value=temperature)

    def get_temperature(self) -> float:
        """
        Get temperature from the controller

        returns
        -------
        temperature:float
            temperature in °C

        raises
        ------
        FurnaceStateException
            if connect method was not called
        """
        if not self._connected:
            raise FurnaceStateException('Connect to furnace controller first!')
        self._logger.info('Requesting temperature from furnace controller.')
        temperature = self._owen_protocol.request_PIC(parameter='pv')
        self._logger.info(f'{temperature = }')
        return temperature

    def set_temperature_control(self, value:bool):
        """
        Turn temperature control ON or OFF.

        parameters
        ----------
        value:bool
            True to turn temperature control ON

        raises
        ------
        FurnaceStateException
            if connect method was not called
        """
        if not self._connected:
            raise FurnaceStateException('Connect to furnace controller first!')
        self._logger.info(f'Setting temperature control to {value}')
        temperature_control = 1 if value else 0
        self._owen_protocol.send_unsigned_byte(parameter='r-s', value=temperature_control)



import multiprocessing
import multiprocessing.connection
import time
import threading

from pycatalicism.furnace.owen_tmp101 import OwenTPM101
from pycatalicism.mass_flow_controller.bronkhorst_f201cv import BronkhorstF201CV
from pycatalicism.plotters.non_blocking_plotter import NonBlockingPlotter

class DataCollectorPlotter(threading.Thread):
    """
    Class for plotting data from activation and activity measurement experiments.
    """

    def __init__(self, furnace_controller:OwenTPM101, mass_flow_controllers:list[BronkhorstF201CV]):
        """
        Initialize base class, instance variables and start non blocking plotter in a separate process.

        parameters
        ----------
        furnace_controller:OwenTPM101
            furnace controller to get information about current temperature
        mass_flow_controllers:list[BronkhorstF201CV]
            list of mass flow controllers to get information about current flow rates
        """
        super().__init__(daemon=False)
        self._furnace_controller = furnace_controller
        self._mfcs = mass_flow_controllers
        self._collector_pipe, self._plotter_pipe = multiprocessing.Pipe()
        self._plotter = NonBlockingPlotter()
        self._plotter_process = multiprocessing.Process(target=self._plotter, args=(self._plotter_pipe,), daemon=False)
        self._plotter_process.start()

    def run(self):
        """
        This method is invoked when thread is started. While thread is running get temperature and flow rate data and send these data through multiprocessing pipe to non blocking plotter. Send None to plotter, when thread is not running.
        """
        self._running = True
        self._start_time = time.time()
        while self._running:
            temperature, flow_rates = self._collect_data()
            self._send_data(temperature, flow_rates)
            time.sleep(10)
        self._send_data(None, None)

    def stop(self):
        """
        Stop this thread.
        """
        self._running = False

    def _collect_data(self) -> tuple[list[float], list[list[float]]]:
        """
        Collect temperature and flow rates data and send these to plotter with time stamps.

        returns
        -------
        temperature:list[float]
            list of time (0th) and temperature(1st) collected from furnace controller
        flow_rates:list[list[float]]
            list of lists of time (0th) and flow rate (1st) data collected from all mass flow controllers
        """
        temp_t = (time.time() - self._start_time) / 60.0
        temp_T = self._furnace_controller.get_temperature()
        temperature = [temp_t, temp_T]
        flow_rates = []
        for mfc in self._mfcs:
            t = (time.time() - self._start_time) / 60.0
            fr = mfc.get_flow_rate()
            flow_rates.append([t, fr])
        return (temperature, flow_rates)

    def _send_data(self, temperature:list[float]|None, flow_rates:list[list[float]]|None):
        """
        Send data to non blocking plotter through pipe.

        parameters
        ----------
        temperature:list[float]|None
            list of time (0th) and temperature (1st) data to send to plotter or None if thread was stopped
        flow_rates:list[list[float]]|None
            list of lists of time (0th) and temperature (1st) data to send to plotter or None if thread was stopped
        """
        self._collector_pipe.send((temperature, flow_rates))



class Activity():
    """
    Wrapper for activity data storage. Not implemented at the moment.
    """

    def __init__(self):
        """
        Initialize activity object. Not implemented at the moment
        """
        raise NotImplementedError()



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


install_requires = [
    "numpy>=1.23.2",
    "matplotlib>=3.5.3",
    "pyserial>=3.5",
    "pymodbus>=2.5.3",
    "bronkhorst-propar>=1.0",
    ]

setuptools.setup(
     name='pycatalicism',
     version='0.2.3',
     author="Denis Leybo",
     author_email="leybodv@gmail.com",
     description="Program controls catalytic activity of materials measurement equipment as well calculations",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/leybodv/pycatalicism",
     packages=setuptools.find_packages(),
     install_requires = install_requires,
     python_requires='>3.10.0',
     entry_points={
                        'console_scripts': [
                                'pycat=pycatalicism.pycat:main',
                        ]},
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
 )



import multiprocessing.connection

import matplotlib.pyplot as plt
import matplotlib
import matplotlib.axes

class NonBlockingPlotter():
    """
    Non blocking plotter that should be run in a separate process due to the incompatibility of matplotlib with multithreading. It will draw data from instance variables every minute after the process was started. Data must be sent over multiprocessing pipe.
    """

    def __init__(self):
        """
        Initialize instance variables.
        """
        self._temp_time = []
        self._temp_temperature = []
        self._fr_times = [[],[],[]]
        self._fr_flow_rates = [[],[],[]]

    def __call__(self, pipe:multiprocessing.connection.Connection):
        """
        Create figure and axeses, setup them, start timer and show canvas.

        parameters
        ----------
        pipe:multiprocessing.connection.Connection
            pipe through which data are get from data collector
        """
        self._pipe = pipe
        self._fig, self._left_ax = plt.subplots()
        self._setup_left_ax(self._left_ax)
        self._right_ax = self._left_ax.twinx()
        self._setup_right_ax(self._right_ax)
        timer = self._fig.canvas.new_timer(interval=60000)
        timer.add_callback(self._call_back)
        timer.start()
        plt.show()

    def _call_back(self) -> bool:
        """
        Collect data through multiprocessing pipe, append them to instance variables and add corresponding dots on plot canvas. This function is called by the matplotlib timer. Data must be sent as iterable with temperature data at 0th position and flow rates data at 1st. Temperature data is a list of floats with time at 0th position and temperature at 1st. Flow rates is a 3-element list, each containing data from different mass flow controllers. Each element is in turn a list with time at 0th position and temperature at 1st.

        returns
        -------
        timer_is_running:bool
            True if timer is needed to be run, False if it must be stopped
        """
        while self._pipe.poll():
            data = self._pipe.recv()
            temperature = data[0]
            flow_rates = data[1]
            if temperature is None or flow_rates is None:
                return False
            else:
                self._temp_time.append(temperature[0])
                self._temp_temperature.append(temperature[1])
                for i in range(3):
                    self._fr_times[i].append(flow_rates[i][0])
                    self._fr_flow_rates[i].append(flow_rates[i][1])
                self._left_ax.clear()
                self._setup_left_ax(self._left_ax)
                self._left_ax.plot(self._temp_time, self._temp_temperature)
                self._right_ax.clear()
                self._setup_right_ax(self._right_ax)
                for i in range(3):
                    self._right_ax.plot(self._fr_times[i], self._fr_flow_rates[i])
        self._fig.canvas.draw()
        return True

    def _setup_left_ax(self, left_ax:matplotlib.axes.Axes):
        """
        Setup left, temperature, axes. Set x and y labels

        parameters
        ----------
        left_ax:matplotlib.axes.Axes
            axes to setup
        """
        left_ax.set_xlabel('Time')
        left_ax.set_ylabel('Temperature')

    def _setup_right_ax(self, right_ax:matplotlib.axes.Axes):
        """
        Setup right, flow rates, axes. Sets y label.

        parameters
        ----------
        right_ax:matplotlib.axes.Axes
            axes to setup
        """
        right_ax.set_ylabel('Flow rate')



from pycatalicism.calc.rawdata import RawData
from pycatalicism.calc.conversion import Conversion
from pycatalicism.calc.selectivity import Selectivity
from pycatalicism.calc.activity import Activity

class Calculator():
    """
    Abstract class to calculate catalyst's activity, selectivity or conversion for different reactions of interest. Concrete classes should override corresponding abstract methods of this class.
    """

    def calculate_conversion(self, input_data:RawData) -> Conversion:
        """
        Calculate conversion vs. temperature data. Abstract method, should be overriden by concrete classes.

        parameters
        ----------
        input_data:RawData
            input data for calculation containing concentrations of reaction participants at different temperatures

        returns
        -------
        conversion:Conversion
            Wrapper of conversion vs. temperature data

        raises
        ------
        exception:NotImplementedError
            if this method is not overriden but is used
        """
        raise NotImplementedError()

    def calculate_selectivity(self, input_data:RawData) -> Selectivity:
        """
        Calculate selectivity vs. temperature data. Abstract method, should be overriden by concrete classes.

        parameters
        ----------
        input_data:RawData
            input data for calculation, containing concentrations of reaction participants at different temperatures

        returns
        -------
        selectivity:Selectivity
            Wrapper of selectivity vs. temperature data

        raises
        ------
        exception:NotImplementedError
            if this method is not overriden but is used
        """
        raise NotImplementedError()

    def calculate_activity(self, input_data:RawData) -> Activity:
        """
        Calculate activity vs. temperature data. Abstract method, should be overriden by concrete classes.

        parameters
        ----------
        input_data:RawData
            input data for calculation, containing concentrations of reaction participants at different temperatures

        returns
        -------
        activity:Activity
            Wrapper of activity vs. temperature data

        raises
        ------
        exception:NotImplementedError
            if this method is not overriden but is used
        """
        raise NotImplementedError()
